use bitboard::*;
use movegen::*;
use game::Game;
use evaluate::*;
use log::info;
use rand::prelude::*;

enum MoveUndo {
    Default {
        piece_type: PieceType,
        before: Bitboard,
        captured_type: PieceType,
        captured_before: Bitboard,
    },
    Castling { rooks_before: Bitboard, king_before: Bitboard },
    Promotion {
        promoted_to: PieceType,
        promoted_to_before: Bitboard,
        pawns_before: Bitboard
    }
}


pub struct SearchControl<'a> {
    pub nodes: usize,
    pub check: &'a mut dyn FnMut() -> bool,
}

pub enum SearchResult {
    Finished(Move, PosValue),
    Mate(Move, i32),
    Cancelled(Option<(Move, PosValue)>),
    Invalid
}


/**
 * searches for moves and returns the best move found plus its value
 */
pub fn search(game: &Game, sc: &mut SearchControl, depth: i32) -> SearchResult {

    if depth == 0 {
        return SearchResult::Invalid;
    }

    let mut moves = generate_legal_moves(game, game.turn);
    //info!("mov list: {:?}", moves.iter().map(|x| x.to_string()).collect::<Vec<String>>());

    let mut rng = rand::thread_rng();

    /*moves.shuffle(&mut rng);

    // return random move
    if moves.len() >= 1 {
        return moves[0];
    }*/

    sort_moves(game, &mut moves);
    
    info!("moves: {:?}", moves.iter().map(|mv| mv.to_string()).collect::<Vec<String>>());

    let mut alpha: PosValue = MIN_VALUE;
    let mut beta: PosValue = MAX_VALUE;

    // use a slight offset for the alpha value in the root node in order to
    // determine possibly multiple good moves
    const ALPHA_OFFSET: PosValue = 50 as _;

    let mut valued_moves: Vec<(Move, PosValue)> = Vec::with_capacity(moves.len());
    let mut cancelled = false;
    for mov in moves {
        let new_game = apply_move(game, mov);
        //info!("searching {}", mov.to_string());
        let (mut val, ret) = negamax(&new_game, sc, -beta, -alpha, depth - 1);
        val = -val;

        if ret {
            //return (Move::default(), 0);
            cancelled = true;
            break;
        }

        if val >= MAX_VALUE {
            // mate in 1
            info!("yay mate!");
            return SearchResult::Mate(mov, 1);
        }

        //info!("searched {} -> {}", mov.to_string(), val);

        if val > alpha {
            alpha = val - ALPHA_OFFSET;
            //info!("updated alpha to {}", alpha);
        }

        valued_moves.push((mov, -val));
    }

    //info!("movvalues: {:?}", valued_moves.iter().map(|mv| mv.0.to_string() + " - " + &mv.1.to_string()).collect::<Vec<String>>());

    valued_moves.sort_by_key(|mv| (mv.1 * 1000.0f32) as i64);

    //info!("best movvalues: {:?}", valued_moves.iter().map(|mv| mv.0.to_string() + " - " + &mv.1.to_string()).collect::<Vec<String>>());

    if valued_moves.len() > 0 {
        let min_val = valued_moves[0].1;
        let best_moves = valued_moves.iter().filter(|mv| mv.1 == min_val).collect::<Vec<&(Move, PosValue)>>();

        //info!("bestmove value {}", -min_val);
        let chosen_mov = best_moves[(rng.next_u64() % best_moves.len() as u64) as usize];
        if cancelled {
            return SearchResult::Cancelled(Some((chosen_mov.0, chosen_mov.1)));
        }
        else {
            return SearchResult::Finished(chosen_mov.0, chosen_mov.1);
        }
    }
    else {
        return SearchResult::Invalid;
    }
}

fn negamax(game: &Game, sc: &mut SearchControl, mut alpha: PosValue, beta: PosValue, depth: i32) -> (PosValue, bool) {
    if depth == 0 {
        return (quiescence_search(game, sc, alpha, beta, 7), false);
        let eval = evaluate(game);
        if eval != 0.0f32 {
            //info!("eval: {}", eval);
        }
        //return eval;
    }

    sc.nodes += 1;
    if sc.nodes % 128 == 0 {
        if (sc.check)() {
            return (0 as _, true);
        }
    }

    let mut best = MIN_VALUE;
    //let mut best_move = Move::default();
    //info!(" -> generate_legal_moves");

    if game.get_piece(KING, game.turn) == 0 { return (MIN_VALUE, false); }

    let moves = generate_legal_moves(game, game.turn);

    if moves.len() == 0 {
        if is_check(game, game.turn) {
            // mate
            return (MIN_VALUE, false);
        }
        else {
            // stalemate
            return (0 as _, false);
        }
    }
    for mov in moves {
        let new_game = apply_move(game, mov);
        let (mut val, ret) = negamax(&new_game, sc, -beta, -alpha, depth - 1);
        val = -val;

        if ret {
            return (alpha, ret)
        }

        if val >= beta {
            return (beta, false);
        }
        if val > alpha {
            alpha = (val as f64 * 0.95) as _;
        }
        //info!(" -> negamaxed {} -> {}", mov.to_string(), depth - 1);
        if val > best {

            best = val;
            //best_move = mov;
        }
    }
    //info!("best alpha {}", best);
    return ((best as f64 * 0.99) as _, false);
}

fn quiescence_search(game: &Game, si: &mut SearchControl, mut alpha: PosValue, beta: PosValue, depth: i32) -> PosValue {
    let val = evaluate(game);
    si.nodes += 1;

    if val >= beta {
        return beta;
    }
    if val > alpha {
        alpha = val;
    }

    if depth <= 0 {
        return alpha;
    }

    if game.get_piece(KING, game.turn) == 0 { return MIN_VALUE; }

    let moves = generate_attacking_moves(game, game.turn);
    for mov in moves {
        let new_game = apply_move(game, mov);
        let val = -quiescence_search(&new_game, si, -beta, -alpha, depth - 1);
        if val >= beta {
            return beta;
        }
        if val > alpha {
            alpha = val;
        }
    }
    return alpha;
}


pub fn apply(game: &mut Game, mov: Move) {
    
}


pub fn undo(game: &mut Game, mov: Move) {
    
}

pub fn apply_move(game: &Game, mov: Move) -> Game {
    let mut new_game = game.clone();

    let side = game.turn;
    let friends = game.get_all_side(side);
    let others = game.get_all_side(!side);
    match mov {
        Move::Default { mov, piece_type: pt } => {

            if pt == KING {
                // invalidate castling rights
                new_game.castling_rights[if game.turn == BLACK { 2 } else { 0 }] = false;
                new_game.castling_rights[if game.turn == BLACK { 3 } else { 1 }] = false;
            }

            if pt == ROOK {
                // invalidate castling rights
                if mov.from == square_from_indices(7, 7) {
                    new_game.castling_rights[0] = false;
                }
                else if mov.from == square_from_indices(0, 7) {
                    new_game.castling_rights[1] = false;
                }
                else if mov.from == square_from_indices(7, 7) {
                    new_game.castling_rights[2] = false;
                }
                else if mov.from == square_from_indices(7, 7) {
                    new_game.castling_rights[3] = false;
                }
            }

            if from_square(mov.to) | others != 0 {
                new_game.apply_mask(!from_square(mov.to));
            }
            let moved_piece = mov.apply_to(new_game.get_piece(pt, side));
            new_game.set_piece(pt, side, moved_piece);
        },
        Move::Castling { side, left } => {
            // invalidate future castling rights
            new_game.castling_rights[if game.turn == BLACK { 2 } else { 0 }] = false;
            new_game.castling_rights[if game.turn == BLACK { 3 } else { 1 }] = false;

            let king = game.get_piece(KING, side);
            let rook = if left {
                game.get_piece(ROOK, side) & (king << 4)
            }
            else {
                game.get_piece(ROOK, side) & (king >> 3)
            };

            let new_king = if left { king << 2 } else { king >> 2 };
            let new_rook = if left { rook >> 3 } else { rook << 2 };

            new_game.set_piece(KING, side, new_king);
            new_game.set_piece(ROOK, side, new_rook);
        },
        Move::EnPassant { mov, beaten } => {
            let pawns = game.pawns(side);
            if let Some(ep) = game.en_passant {
                *new_game.get_piece_mut(PAWN, side) &= !from_square(mov.from);
                *new_game.get_piece_mut(PAWN, side) |= from_square(mov.to);
                *new_game.get_piece_mut(PAWN, !side) |= from_square(beaten);
            }
        },
        Move::Promotion { mov, promote_to } => {
            if from_square(mov.to) | others != 0 {
                new_game.apply_mask(!from_square(mov.to));
            }
            new_game.apply_mask(!from_square(mov.from));

            match promote_to {
                QUEEN => { *new_game.queens_mut(side) |= from_square(mov.to); }
                ROOK => { *new_game.rooks_mut(side) |= from_square(mov.to); }
                BISHOP => { *new_game.bishops_mut(side) |= from_square(mov.to); }
                KNIGHT => { *new_game.knights_mut(side) |= from_square(mov.to); }
                _ => {
                    info!("internal error");
                }
            }
        },
    }

    new_game.turn = !new_game.turn;



    return new_game;
}
